#from flask            import Flask, jsonify, request
#from flask_restful   import Resource, Api
#from importlib       import import_module
#from importlib.util  import find_spec, LazyLoader, module_from_spec, spec_from_file_location
#from inspect         import getmembers, getmodulename, isclass
#from sys             import modules, path
#from tempfile        import NamedTemporaryFile
from ratelimit        import limits, sleep_and_retry
from requests         import post
from teamhack_db.sql  import insert
from teamhack_db.util import get_name, get_record_type
from            .sql  import *
#from             sql  import *
from            .util import diff
#from            util import diff

def portscan(queue):
  print(f'portscan({queue})')
  nmap      = 'http://0.0.0.0:55432/upload'
  response  = post(nmap,      files={'file': queue})
  if response.status_code != 200: return response.text, response.status_code

  import_db = 'http://0.0.0.0:65432/upload'
  response  = post(import_db, files={'file': response.text})
  return response.text, response.status_code

def subdomains(queue):
  #print(f'subdomains({queue})')
  pass
def vhosts(queue):
  #print(f'vhosts({queue})')
  pass
def subdirectories(queue):
  #print(f'subdirectories({queue})')
  pass
def credentials(queue):
  #print(f'credentials({queue})')
  pass
def flags(queue):
  #print(f'flags({queue})')
  pass

@sleep_and_retry
@limits(calls=3, period=180)
def loop(dns=None, msf=None, sdn=None, *args, **kwargs):
  inbound  = select_dns(dns)
  print(f'inbound: {inbound}')
  inbound  = [k[0] for k in inbound]
  print(f'inbound: {inbound}')

  psq      =     portscan_queue(inbound, msf) # msfcli   db_nmap psq
  print(f'psq: {psq}')
  text, code = portscan(psq) # batch process
  print(f'text: {text}')
  if code != 200: raise Exception(f'code: {code}')

  #svq      =      service_queue(inbound, sdn) # TODO should be populated by db_nmap ?

  sdq      =    subdomain_queue(inbound, sdn) # gobuster dns     sdq
  print(f'sdq: {sdq}')
  #sdq      =    subdomain_queue(psq, sdn) # gobuster dns     sdq
  subdomains(sdq) # sequential process

  vhq      =        vhost_queue(inbound, sdn) # gobuster vhost   shq
  #vhq      =        vhost_queue(sdq, sdn) # gobuster vhost   shq
  vhosts(vhq) # sequential process

  fpq      = subdirectory_queue(inbound, sdn) # gobuster dir     fpq
  #fpq      = subdirectory_queue(vhq, sdn) # gobuster dir     fpq
  subdirectories(fpq) # sequential process

  crq      =   credential_queue(inbound, sdn) # hydra
  credentials(crq) # sequential process

  fgq      =         flag_queue(inbound, sdn) # ssh linpeas, pspy
  flags(fgq)

  # TODO
  # - services found
  #   - scrape web for CVE re: service versions
  #   - 80, 443
  #     - begin vhost        scan
  #     - begin subdirectory scan
  #     - recursively download website
  #       - static analysis
  #       - bulk_extractor => generate wordlists
  #       - cewl
  #       - cupp
  #   - sqlmap
  #   - hydra/ncrack/medusa
  #   - db_autopwn
  pass



#def create_app(conn):
#  app = Flask(__name__)
#  #api = Api(app)
#
#  def dispatch(data, hostname_recordtype_cb, hostname_cb, ip_cb):
#    if 'host' in data and 'type' in data:
#      host = data['host']
#      host = get_name(host)
#      rt   = data['type']
#      rt   = get_record_type(rt)
#      ret  = hostname_recordtype_cb(conn, host, rt)
#      return ret
#    if 'host' in data and 'type' not in data:
#      host = data['host']
#      host = get_name(host)
#      ret  = hostname_cb(conn, host)
#      return ret
#    if 'inet' in data:
#      addr = data['inet']
#      ret  = ip_cb(conn, addr)
#      return ret
#    return '', 404
#
#  @app.route('/create', methods=['POST'])
#  def add():
#    data = request.get_json(force=True)
#    if 'host' not in data: return '', 404
#    host = data['host']
#    host = get_name(host)
#    if 'type' not in data: return '', 404
#    rt   = data['type']
#    rt   = get_record_type(rt)
#    if 'inet' not in data: return '', 404
#    addr = data['inet']
#    insert(conn, host, rt, addr)
#    return '', 204
#
#  @app.route('/retrieve', methods=['POST'])
#  def retrieve():
#    data = request.get_json(force=True)
#    return dispatch(data, select_hostname_recordtype, select_hostname, select_ip)
#
#  @app.route('/update', methods=['POST'])
#  def update():
#    # TODO
#    return '', 404
#
#  @app.route('/delete', methods=['POST'])
#  def delete():
#    data = request.get_json(force=True)
#    return dispatch(data, drop_row_hostname_recordtype, drop_row_hostname, drop_row_ip)
#
#  return app

def start_server(host="0.0.0.0", port=6000, dns=None, msf=None, sdn=None, *args, **kwargs):
  #psd = portscan_daemon()
  #while(True): loop(dns, msf, sdn, psd, **kwargs)
  while(True): loop(dns, msf, sdn, **kwargs)
  #app = create_app(dns=dns, msf=msf, **kwargs)
  #app.run(debug=True, host=host, port=port, *args)




from psycopg2           import connect
from teamhack_db.conf   import config
from            .server import start_server
#from             server import start_server
from             .sql   import create_tables
#from              sql   import create_tables

if __name__ == '__main__':
  conns        = dict()
  for db in ('dns', 'msf', 'sdn'): # TODO how to read all sections
    conns[db]  = connect(**config(section=db))

  create_tables(conns['dns'], conns['sdn'])

  start_server(**conns)




from teamhack_db.sql  import create_table, insert, execute, select
from teamhack_db.util import get_name, get_record_type
from            .util import diff
#from             util import diff

def create_table_sdn  (conn): return execute(conn, """
  CREATE TABLE IF NOT EXISTS subdomains  (
    id        SERIAL        PRIMARY KEY,
    subdomain VARCHAR(255)  NOT NULL,
    host      INET          NOT NULL,
    UNIQUE(subdomain, host)
  )
""")

# TODO port ?
def create_table_vh   (conn): return execute(conn, """
  CREATE TABLE IF NOT EXISTS vhosts      (
    id        SERIAL        PRIMARY KEY,
    vhost     VARCHAR(255)  NOT NULL,
    host      INET          NOT NULL,
    UNIQUE(vhost, host)
  )
""")

# TODO port ?
def create_table_fp   (conn): return execute(conn, """
  CREATE TABLE IF NOT EXISTS subdirs     (
    id        SERIAL        PRIMARY KEY,
    vhostid   INT           NOT NULL REFERENCES vhosts(id),
    path      VARCHAR(255)  NOT NULL,
    UNIQUE(vhostid, path)
  )
""")

def create_table_srv  (conn): return execute(conn, """
  CREATE TABLE IF NOT EXISTS services    (
    id        SERIAL        PRIMARY KEY,
    subdnid   INT           NOT NULL REFERENCES subdomains(id),
    port      INT2          NOT NULL,
    UNIQUE(subdnid, port)
  )
""")

def create_table_creds(conn): return execute(conn, """
  CREATE TABLE IF NOT EXISTS credentials (
    id        SERIAL        PRIMARY KEY,
    serviceid INT           NOT NULL REFERENCES services(id),
    username  VARCHAR(255),
    password  VARCHAR(255),
    UNIQUE(serviceid, username)
  )
""")

def create_table_flags(conn): return execute(conn, """
  CREATE TABLE IF NOT EXISTS flags       (
    id        SERIAL        PRIMARY KEY,
    host      INET          NOT NULL,
    flag      VARCHAR(255)  NOT NULL,
    path      VARCHAR(255)  NOT NULL,
    isroot    BOOLEAN       NOT NULL,
    UNIQUE(host, path),
    UNIQUE(host, isroot)
  )
""")

def drop_table_sdn  (conn): return execute(conn, """DROP TABLE IF EXISTS subdomains""")
def drop_table_vh   (conn): return execute(conn, """DROP TABLE IF EXISTS vhosts""")
def drop_table_fp   (conn): return execute(conn, """DROP TABLE IF EXISTS subdirs""")
def drop_table_srv  (conn): return execute(conn, """DROP TABLE IF EXISTS services""")
def drop_table_creds(conn): return execute(conn, """DROP TABLE IF EXISTS credentials""")
def drop_table_flags(conn): return execute(conn, """DROP TABLE IF EXISTS flags""")

def drop_tables(conn):
  drop_table_flags(conn)
  drop_table_creds(conn)
  drop_table_srv  (conn)
  drop_table_fp   (conn)
  drop_table_vh   (conn)
  drop_table_sdn  (conn)

def create_tables(dns, sdn):
  drop_tables(sdn) # TODO delete

  create_table(dns)
  dns.commit()
  create_table_sdn  (sdn)
  create_table_vh   (sdn)
  create_table_fp   (sdn)
  create_table_srv  (sdn)
  create_table_creds(sdn)
  create_table_flags(sdn)
  sdn.commit()

def select_dns  (conn):    return select(conn, """
  SELECT ip
  FROM hosts
""")
#def select_msf(conn, q): return select(conn, """
#  SELECT address           FROM hosts
#  WHERE address IN %s
#""", (tuple(q),))
def select_msf (conn, q):
  print("select_msf(q=%s)" % (q,))
  return select(conn, """
  SELECT address
  FROM hosts
  WHERE address IN %s
""", (tuple(q),))
def select_sdn (conn, q): return select(conn, """
  SELECT subdomain, host
  FROM subdomains
  WHERE host    IN %s
""", (tuple(q),))
def select_vh  (conn, q): return select(conn, """
  SELECT vhost,     host
  FROM vhosts
  WHERE host    IN %s
""", (tuple(q),))
def select_fp  (conn, q): return select(conn, """
  SELECT vhost,     host, path
  FROM subdirs
  JOIN vhosts ON vhostid = vhosts.id
  WHERE host   IN %s
""", (tuple(q),))
def select_srv (conn, q): return select(conn, """
  SELECT vhost,     host, port
  FROM services
  JOIN vhosts ON vhostid = vhosts.id
  WHERE host   IN %s
""", (tuple(q),))
def select_cred(conn, q): return select(conn, """
  SELECT subdomain, host, port, username, password
  FROM credentials
  JOIN services   ON serviceid =   services.id
  JOIN subdomains ON   subdnid = subdomains.id
  WHERE host   IN %s
""", (tuple(q),))
def select_flag(conn, q): return select(conn, """
  SELECT            host, path, flag, isroot
  FROM flags
  WHERE host   IN %s
""", (tuple(q),))

def get_queue(inbound, get_outbound, db):
  outbound = get_outbound(db,       inbound)
  print(f'outbound: {outbound}')
  ret      = diff        (inbound, outbound)
  #print(f'A diff: {diff( inbound, outbound)}')
  #print(f'B diff: {diff(outbound,  inbound)}')
  print(f'ret: {ret}')
  return ret

def     portscan_queue(inbound, msf): return get_queue(inbound, select_msf,  msf)
def    subdomain_queue(inbound, sdn): return get_queue(inbound, select_sdn,  sdn)
def        vhost_queue(inbound, sdn): return get_queue(inbound, select_vh,   sdn)
def subdirectory_queue(inbound, sdn): return get_queue(inbound, select_fp,   sdn)
def   credential_queue(inbound, sdn): return get_queue(inbound, select_cred, sdn)
def         flag_queue(inbound, sdn): return get_queue(inbound, select_flag, sdn)

